class lib:
    @staticmethod
    def start_instrumentation() -> None: ...
    @staticmethod
    def stop_instrumentation() -> None: ...
    @staticmethod
    def dump_stats() -> None: ...
    @staticmethod
    def dump_stats_at(trigger: bytes) -> None: ...
    @staticmethod
    def zero_stats() -> None: ...
    @staticmethod
    def toggle_collect() -> None: ...
